/***************************************************************************/
/* C++ Code to estimate variance components and effects using a gibbs      */
/* sampler in the context of an animal model. The program borrows the      */
/* 'MeuwissenLuo_1992_F.cpp' code to generate Ainv and the same rules      */
/* which takes in one argument which contains all of the parameters.       */
/* Use it at your own discression!!                                        */
/***************************************************************************/
/* To compile type: 'make'                                                 */
/* To run type: ./gibbs parameterfile                                      */
/***************************************************************************/
#include <iostream>
#include <vector>
#include <string>
#include <fstream>
#include <algorithm>
#include <math.h>
#include <iomanip>
#include <unordered_map>
#include <set>
#include <tuple>
#include <sstream>
#include <random>
#include <Eigen/Dense>
#include <Eigen/Sparse>
#include <Eigen/Core>
#include "Parameters.h"
#include "Pedigree.h"

using namespace std;
/*****************************************************/
/*****   Functions used in Gibbs Sampling Program ****/
/*****************************************************/
void read_generate_parameters(parameters &Parameters, string parameterfile, ostream& logfile); /* Reads in Parameters and declares parameter values */
void read_sort_ainv(vector <Pedigree_Class> &pedigree, vector <tuple<int,int,double>> &sprelationshipinv, parameters &Parameters,vector<int> &nonzerorows);
/*****************************************/
/*****   Begin Gibbs Sampling Program ****/
/*****************************************/
int main(int argc, char* argv[])
{
    using Eigen::SparseMatrix; using Eigen::MatrixXd; using Eigen::VectorXd; typedef Eigen::Triplet<double> T;
    time_t full_begin_time = time(0);
    cout << endl;
    cout <<"+-----+-----+-----+-----+-----+-----+-----+------+-----+-----+" << endl;
    cout <<"|                                                            |" << endl;
    cout <<"|  Gibbs Sampler to Estimate Variance Components and Effects |" << endl;
    cout <<"|  Author: Jeremy T. Howard                                  |" << endl;
    cout <<"|                                                            |" << endl;
    cout <<"+-----+-----+-----+-----+-----+-----+-----+------+-----+-----+" << endl;
    cout <<"| This program is free software: you can redistribute it     |" << endl;
    cout <<"| and/or modify it under the terms of the GNU General Public |" << endl;
    cout <<"| License as published by the Free Software Foundation,      |" << endl;
    cout <<"| either version 3 of the License, or (at your option) any   |" << endl;
    cout <<"| later version.                                             |" << endl;
    cout <<"+-----+-----+-----+-----+-----+-----+-----+------+-----+-----+" <<endl;
    cout << " - Reading in Parameter File." << endl;
    if(argc != 2){cout << "Need to give Parameter File Name!!" << endl; exit (EXIT_FAILURE);}
    string paramterfile = argv[1];
    parameters Parameters;       /* Initialize parameter class */
    fstream checklog; checklog.open("logfile", std::fstream::out | std::fstream::trunc); checklog.close(); /* Deletes log file Previous Simulation */
    std::ofstream logfile("logfile", std::ios_base::app | std::ios_base::out);               /* open log file to output verbage throughout code */
    read_generate_parameters(Parameters,paramterfile,logfile);      /* Figure out parameters and model */
    /** Update number of threads for Eigen Multi-threading */
    omp_set_num_threads(Parameters.getthreads());
    Eigen::setNbThreads(Parameters.getthreads());
    int thread = Eigen::nbThreads();
    vector < Pedigree_Class > pedigree;                             /* Save information about animal in vector of "Pedigree_Class" */
    vector < tuple <int,int,double> > sprelationshipinv;            /* Store Ainv in sparse format */
    vector <int> nonzerorows;                                       /* Numbero of non-zero's by row */
    /******************************************************************/
    /* Read, sort and generate Ainverse to use when solving equations */
    /******************************************************************/
    time_t Ainv_begin_time = time(0);
    read_sort_ainv(pedigree,sprelationshipinv,Parameters,nonzerorows);
    time_t Ainv_end_time = time(0);
    logfile << "===============================================" << endl;
    logfile << "==        Begin Setting Up LHS and RHS       ==" << endl;
    logfile << "===============================================" << endl;
    logfile << " - Number of animals in pedigree file: "<<pedigree.size()<<"."<<endl;
    SparseMatrix<double> Ainv(pedigree.size(),pedigree.size());
    Ainv.reserve(nonzerorows);
    vector<T> tripletList; tripletList.reserve(sprelationshipinv.size());
    for(int i = 0; i < sprelationshipinv.size(); i++){tripletList.push_back(T(get<0>(sprelationshipinv[i]),get<1>(sprelationshipinv[i]),get<2>(sprelationshipinv[i])));}
    Ainv.setFromTriplets(tripletList.begin(),tripletList.end()); sprelationshipinv.clear();
    Ainv.makeCompressed();
    logfile << "    - Constructed Inverse of A (Took: "<<difftime(Ainv_end_time,Ainv_begin_time)<<" seconds)"<<endl<<endl;
    /******************************************************************/
    /* Read in phenotypes and generate observation by specific column */
    /******************************************************************/
    vector<vector<double>> pheno;
    vector<vector<double>> residual;
    vector<vector<string>> FIXED_CLASS;
    vector<vector<double>> FIXED_COV;
    vector<vector<string>> RANDOM_CLASS;
    ifstream infile1; string line; int linenumber = 0;
    infile1.open((Parameters.getdatafile()).c_str());
    if(infile1.fail()){cout << "Error Opening Data File \n"; exit (EXIT_FAILURE);}
    while (getline(infile1,line))
    {
        //cout << "'" << line << "'" << endl;
        vector <string> variables;
        while(1)
        {
            size_t pos = line.find(" ",0);
            if(pos > 0){variables.push_back(line.substr(0,pos)); line.erase(0,pos+1);}  /* hasn't reached last one yet */
            if(pos == 0){line.erase(0,pos+1);}                                          /* blank lines at end of each row */
            if(pos == std::string::npos){break;}                                        /* has reached last one so now save last one and kill while loop */
        }
        //for(int i = 0; i < variables.size(); i++){cout<<variables[i]<<endl;}
        /***************************/
        /* Fill pheno 2-D vector < */
        /***************************/
        vector <double> temp; vector <double> tempres; vector <string> tempa;
        for(int i = 0; i < (Parameters.get_trait_col()).size(); i++)
        {
            temp.push_back(atof((variables[(Parameters.get_trait_col())[i]-1]).c_str())); tempres.push_back(0.0);
        }
        pheno.push_back(temp); residual.push_back(tempres); temp.clear(); tempres.clear();
        /*******************************/
        /* Fill FIXED_CLASS 2-D vector */
        /*******************************/
        for(int i = 0; i < (Parameters.get_fixed_type()).size(); i++)
        {
            if((Parameters.get_fixed_type())[i] == "class" || (Parameters.get_fixed_type())[i] == "int")
            {
                tempa.push_back(variables[(Parameters.get_fixed_col())[i]-1]);
            }
        }
        FIXED_CLASS.push_back(tempa); tempa.clear();
        /*****************************/
        /* Fill FIXED_COV 2-D vector */
        /*****************************/
        for(int i = 0; i < (Parameters.get_fixed_type()).size(); i++)
        {
            if((Parameters.get_fixed_type())[i] == "cov"){temp.push_back(atof((variables[(Parameters.get_fixed_col())[i]-1]).c_str()));}
        }
        FIXED_COV.push_back(temp); temp.clear();
        /*******************************/
        /* Fill RANDOM_CLASS 2-D vector */
        /*******************************/
        for(int i = 0; i < (Parameters.get_random_col()).size(); i++)
        {
            tempa.push_back(variables[(Parameters.get_random_col())[i]-1]);
        }
        RANDOM_CLASS.push_back(tempa); tempa.clear();
        //cout << pheno.size() << " " << pheno[0].size() << " " << pheno[0][0] << endl;
        //cout << FIXED_CLASS.size() << " " << FIXED_CLASS[0].size() << " " << FIXED_CLASS[0][0] << endl;
        //cout << FIXED_COV.size() << " " << FIXED_COV[0].size() << endl;
        //cout << RANDOM_CLASS.size() << " " << RANDOM_CLASS[0].size() << " " << RANDOM_CLASS[0][0] << endl;
    }
    //cout << pheno.size() << " " << FIXED_CLASS[0].size() << " " << FIXED_COV[0].size() << " " << RANDOM_CLASS[0].size() << endl;
    /** figures out start and stop portion to estimate LHS when changing A **/
    int startanim = FIXED_CLASS[0].size() + FIXED_COV[0].size();
    int endanim = FIXED_CLASS[0].size() + FIXED_COV[0].size() + RANDOM_CLASS.size() -1;
    /*********************************************************************************************/
    /* Generate hash_map to figure where each fixed and random element is located in LHS and RHS */
    /*********************************************************************************************/
    vector < string > group;
    unordered_map <string,int> LHS_RHS_linker;                  /* Hash-table data structure for faster lookup */
    std:unordered_map<string,int>::const_iterator got;          /* Search if key exists */
    string tempstring;
    for(int i = 0; i < FIXED_CLASS[0].size(); i++)
    {
        for(int j = 0; j < FIXED_CLASS.size(); j++)
        {
            tempstring = "CLASS_"+to_string(i+1)+"_"+FIXED_CLASS[j][i];
            got = LHS_RHS_linker.find(tempstring);
            if(got == LHS_RHS_linker.end())
            {
                LHS_RHS_linker.insert({tempstring,LHS_RHS_linker.size()});
                group.push_back(tempstring);
            }
        }
    }
    for(int i = 0; i < FIXED_COV[0].size(); i++)
    {
        for(int j = 0; j < FIXED_COV.size(); j++)
        {
            tempstring = "COV_"+to_string(i+1);
            got = LHS_RHS_linker.find(tempstring);
            if(got == LHS_RHS_linker.end())
            {
                LHS_RHS_linker.insert({tempstring,LHS_RHS_linker.size()});
                group.push_back(tempstring);
            }
        }
    }
    for(int i = 0; i < RANDOM_CLASS[0].size(); i++)
    {
        if((Parameters.get_random_type())[i] == "AINV")
        {
            /* Loop through sorted pedigree file to generate the correct order that compares with AINV matrix */
            for(int j = 0; j < pedigree.size(); j++)
            {
                tempstring = "RANDOM_"+to_string(i+1)+"_"+pedigree[j].getanim();
                got = LHS_RHS_linker.find(tempstring);
                if(got == LHS_RHS_linker.end())
                {
                    LHS_RHS_linker.insert({tempstring,LHS_RHS_linker.size()});
                    group.push_back(tempstring);
                }
            }
        } else{
            cout << "Shouldn't Be Here" << endl; exit (EXIT_FAILURE);
            for(int j = 0; j < RANDOM_CLASS.size(); j++)
            {
                tempstring = "RANDOM_"+to_string(i+1)+"_"+RANDOM_CLASS[j][i];
                got = LHS_RHS_linker.find(tempstring);
                if(got == LHS_RHS_linker.end())
                {
                    LHS_RHS_linker.insert({tempstring,LHS_RHS_linker.size()});
                    group.push_back(tempstring);
                }
            }
            
        }
    }
    //for (auto it = LHS_RHS_linker.begin(); it != LHS_RHS_linker.end(); ++it){cout <<" "<< it->first << ":" << it->second << "\t";}
    //cout << endl;
    MatrixXd LHS(LHS_RHS_linker.size(),LHS_RHS_linker.size()); LHS.setZero(LHS_RHS_linker.size(),LHS_RHS_linker.size());
    VectorXd RHS(LHS_RHS_linker.size()); RHS.setZero(LHS_RHS_linker.size());
    VectorXd b(LHS_RHS_linker.size()); b.setZero(LHS_RHS_linker.size());
    VectorXd b_add(LHS_RHS_linker.size()); b_add.setZero(pedigree.size());
    vector < double > posteriormeansolutions(b.size(),0.0);             /* stores posterior means */
    for(int i = 0; i < pheno.size(); i++)
    {
        vector < int > add1; vector < double > addcoef;
        for(int j = 0; j < FIXED_CLASS[0].size(); j++)
        {
            tempstring = "CLASS_"+to_string(j+1)+"_"+FIXED_CLASS[i][j];
            LHS(LHS_RHS_linker[tempstring],LHS_RHS_linker[tempstring]) += 1;
            RHS(LHS_RHS_linker[tempstring]) += pheno[i][0];
            add1.push_back(LHS_RHS_linker[tempstring]);
        }
        for(int j = 0; j < FIXED_COV[0].size(); j++)
        {
            tempstring = "COV_"+to_string(i+1);
            cout << "Shouldn't be here" << endl; exit (EXIT_FAILURE);
        }
        for(int j = 0; j < RANDOM_CLASS[0].size(); j++)
        {
            tempstring = "RANDOM_"+to_string(j+1)+"_"+RANDOM_CLASS[i][j];
            LHS(LHS_RHS_linker[tempstring],LHS_RHS_linker[tempstring]) += 1;
            RHS(LHS_RHS_linker[tempstring]) += pheno[i][0];
            for(int k = 0; k < add1.size(); k++)
            {
                LHS(add1[k],LHS_RHS_linker[tempstring]) += 1;
                LHS(LHS_RHS_linker[tempstring],add1[k]) += 1;
            }
            add1.push_back(LHS_RHS_linker[tempstring]);
        }
    }
    cout << " - Setup LHS and RHS of Mixed Model Equations." << endl;
    //for(int i = 0; i < 15; i++)
    //{
    //    for(int j = 0; j < 15; j++){cout << LHS(i,j) << " ";}
    //    cout << endl;
    //}
    //cout << endl << endl;
    //for(int i = 0; i < 10; i++)
    //{
    //    cout << RHS(i) << " " << endl;
    //}
    cout << " - Begin Gibbs Sampling:" << endl;
    time_t begin_time = time(0);
    /************************/
    /* Begin Gibbs Sampling */
    /************************/
    int v_e = 10; // flat prior for variance is - 2
    int v_a = 10; // flat prior for variance is - 2
    double S_e = (Parameters.get_rresidual_start())[0];
    double S_a = (Parameters.get_random_start())[0];
    int n = pheno.size();
    MatrixXd sigma_e(1,1); sigma_e(0,0) = S_e;
    MatrixXd sigma_a(1,1); sigma_a(0,0) = S_a;
    MatrixXd LHS_Upd(LHS_RHS_linker.size(),LHS_RHS_linker.size());
    mt19937 gen(1200);                                                  /* Generate random number to start with */
    chi_squared_distribution<double> distr_chisq_e(n+v_e);                              /* sigma^2e chi square*/
    chi_squared_distribution<double> distr_chisq_a(pedigree.size()+v_a);              /* sigma^2e chi square*/
    /***********************/
    /* Temporary Variables */
    /***********************/
    double ete, meana, sda;
    int currentsavediteration = 1; double postermeananim = 0.0; double postermeanresidual = 0.0;
    
    MatrixXd mean; MatrixXd sd;
    for(int iter = 0; iter < Parameters.getsamples(); iter++)
    {
        /** Update LHS with appropriate updated variance components */
        LHS_Upd = LHS; LHS_Upd.block(startanim,startanim,endanim,endanim) += (Ainv * (sigma_e(0,0) / sigma_a(0,0)));
        /* update each beta */
        for(int i = 0; i < LHS_RHS_linker.size(); i++)
        {
            meana = (RHS(i) - ((LHS_Upd.row(i)*b) - (LHS_Upd(i,i)*b(i)))) / double(LHS_Upd(i,i));
            sda = sqrt((sigma_e(0,0) / double(LHS_Upd(i,i))));
            std::normal_distribution<double> normal(meana,sda);
            b(i) = normal(gen);
            if(i > 0){b_add(i-1) = b(i);}
        }
        /********************/
        /* Update Residuals */
        /********************/
        ete = 0.0;
        for(int i = 0; i < pheno.size(); i++)
        {
            //cout << pheno[i][0] << " " << residual[i][0] << endl;
            residual[i][0] = 0.0;
            for(int j = 0; j < FIXED_CLASS[0].size(); j++)
            {
                tempstring = "CLASS_"+to_string(j+1)+"_"+FIXED_CLASS[i][j]; residual[i][0] += b(LHS_RHS_linker[tempstring]);
            }
            for(int j = 0; j < FIXED_COV[0].size(); j++)
            {
                tempstring = "COV_"+to_string(j+1);
                cout << "Shouldn't be here" << endl; exit (EXIT_FAILURE);
            }
            for(int j = 0; j < RANDOM_CLASS[0].size(); j++)
            {
                tempstring = "RANDOM_"+to_string(j+1)+"_"+RANDOM_CLASS[i][j]; residual[i][0] += b(LHS_RHS_linker[tempstring]);
            }
            ete += ((pheno[i][0] - residual[i][0]) * (pheno[i][0] - residual[i][0]));
        }
        sigma_e(0,0) = (ete + (v_e*S_e)) / double(distr_chisq_e(gen));
        /****************************/
        /* Update Additive Variance */
        /****************************/
        mean = b_add.transpose() * Ainv * b_add;
        sigma_a(0,0) = (double(mean(0,0)) + (v_a*S_a)) / double(distr_chisq_a(gen));
        if(iter % 100 == 0 || iter == (Parameters.getsamples()-1))
        {
            cout<<"   - "<<iter+1<<": "<<sigma_e(0,0)<<" "<<sigma_a(0,0)<<endl;
        }
        if((iter+1) > Parameters.getburnin() && (iter+1) % Parameters.getthin() == 0)
        {
            postermeananim = ((postermeananim*(currentsavediteration-1))+sigma_a(0,0)) / double(currentsavediteration);
            postermeanresidual = ((postermeanresidual*(currentsavediteration-1))+sigma_e(0,0)) / double(currentsavediteration);
            for(int i = 0; i < LHS_RHS_linker.size(); i++)
            {
                posteriormeansolutions[i] = ((posteriormeansolutions[i]*(currentsavediteration-1))+b(i)) / double(currentsavediteration);
            }
            currentsavediteration += 1;
        }
    }
    //SaveSamples.close();
    cout << " - Finished Sampling." << endl;
    cout << "   - Posterior Mean Animal Variance: " << postermeananim << endl;
    cout << "   - Posterior Mean Residual Variance: " << postermeanresidual << endl << endl;
    cout << " - Begin to Save Results" << endl;
    cout << "   - Output Posterior Means for for each parameter: " << endl;
    /* If file exists overwrite */
    fstream checksolutions;
    checksolutions.open(Parameters.getouteffectfile().c_str(), std::fstream::out | std::fstream::trunc);
    checksolutions.close();
    std::ofstream output1(Parameters.getouteffectfile().c_str(), std::ios_base::app | std::ios_base::out);
    for(int i = 0; i < posteriormeansolutions.size(); i++)
    {
        output1 << group[i] << " " << posteriormeansolutions[i] << endl;
    }
    cout << "   - Output Posterior Means for Variance Components:" << endl << endl;
    fstream checkvariance;
    checkvariance.open(Parameters.getoutvariancefile().c_str(), std::fstream::out | std::fstream::trunc);
    checkvariance.close();
    std::ofstream output2(Parameters.getoutvariancefile().c_str(), std::ios_base::app | std::ios_base::out);
    output2 <<"Animal Variance: " << postermeananim << endl;
    output2 <<"Residual Variance: " << postermeanresidual << endl;
    time_t end_time = time(0);
    cout << " - Finished Gibbs Program (Took " << difftime(end_time,begin_time) << " Seconds)." << endl << endl;
}
